/*
 * Copyright (c) 2003-2011 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "network.h"
#include "serial.h"

/* Maximum number of dest. stations */
#define MAX_DST 10

static int nDst;			/* number of destinations */
static int verbose = 0;
static int cat = 0;
extern int optind;

static void
usage(const char *progname)
{
	fprintf(stderr, "sends data received on serial port to the specified hosts, UDP port %d\n", GPS_PORT);
	fprintf(stderr, "the GPS device must be configured to send data\n");
	fprintf(stderr, "at 57600 bauds, 8bits, no parity, one bit stop\n");
	fprintf(stderr, "usage: %s [-v][-c] serial_device host [host]...\n",
	    progname);
	fprintf(stderr, "  -v (verbose) print to the screen the sent data\n");
	fprintf(stderr, "  -c (cat) send to stdout the raw data\n");

	exit(1);
}

static GPS_SEND_STR *
setupSerial(const char *device)
{
	GPS_SEND_STR *to;

	to = (GPS_SEND_STR *)malloc(sizeof(GPS_SEND_STR));
	if (to == NULL) {
		fprintf(stderr, "%s:%d: cannot allocate memory\n", 
		    __FILE__, __LINE__);
		return NULL;
	}
	to->fd = openSerial(device);
	if (to->fd < 0) {
		fprintf(stderr, "%s:%d: openSerial: %s\n", __FILE__, __LINE__, 
		    strerror(errno));
		free(to);
		return NULL;
	}
	to->len = 0;
	return to;
}

/*
 * wrapper around setupToSocket, for error recovery
 */
static GPS_SEND_STR *
try_reconnect(const char *hostname, u_int16_t port)
{
	GPS_SEND_STR *tmp;
	size_t len;
	
	if (hostname[0] == '/')
		return setupSerial(hostname);
		
	if (gethostbyname(hostname) == NULL)
		return NULL;

	/* to avoid error msg flooding if it fails in setupToSocket */
	tmp = setupToSocket(hostname, port);
	if (tmp == NULL)
		return NULL;
	len = sendto(tmp->fd, NULL, 0, 0,
	    (struct sockaddr *)&(tmp->dst), tmp->len);
	if (len != 0) {
		/* to avoid error msg flooding if it fails when sending data */
		free(tmp);
		return NULL;
	}
	fprintf(stderr, "%s:%d: reopened socket "
	    "for %s\n", __FILE__, __LINE__, hostname);
	return tmp;
}

int
main(int argc, char *argv[])
{
	int d, i, fd, ch;
	size_t  n, nSent;
	GPS_SEND_STR *to[MAX_DST];
	char *prog = argv[0];
	char buf[1024];

	while ((ch = getopt(argc, argv, "cv")) != -1) {
		switch (ch) {
		case 'v':
			verbose = 1;
			break;
		case 'c':
			cat = 1;
			break;
		case '?':
		default:
			usage(argv[0]);
			;
		} /* switch */
	}
	argc -= optind;
	argv += optind;

	if (argc < 2) {
		usage(prog);
	}
	fd = openSerial(argv[0]);
	if (fd == -1) {
		exit(1);
	}
	nDst = argc - 1;
	if (nDst > MAX_DST) {
		fprintf(stderr, "%s: too many destinations\n", prog);
		nDst = MAX_DST;
	}
	for (d = 0; d < nDst; d++) {
		if (argv[d+1][0] == '/') 
			to[d] = setupSerial(argv[d+1]);
		else
			to[d] = setupToSocket(argv[d+1], GPS_PORT);
		if (to[d] == NULL) {
			exit(1);
		}
	} /* for */
	while (1) {
		n = readSerial(fd, buf, sizeof(buf));
		if (n < 0) {
			fprintf(stderr, "%s:%d: readSerial: %s\n",
			    __FILE__, __LINE__, strerror(errno));
			break;
		}
		if (buf[0] != 0x66 && buf[0] != 0x59) {
			if (verbose) 
				printf("\n dropping bad frame\n");
			continue;
		}
		if (verbose) {
			printf("\nreceived %lu bytes from gps\n", n);
			for (i = 0; i < n; i++) {
				printf("%#02x ", buf[i]);
			}
			printf("\n");
		}
		if (cat) {
			fwrite(buf, n, 1, stdout);
			fflush(stdout);
		}
		/* send data to each host */
		for (d = 0; d < nDst; d++) {
			/*
			 * Try to open the socket again if the connection 
			 * was lost.  
			 * It happens when the machine running this program 
			 * loses its network connection.
			 */
			if (to[d] == NULL)
				to[d] = try_reconnect(argv[d+1], GPS_PORT);

			if (to[d] == NULL) /* failed above */
				continue;
			if (to[d]->len == 0)
				nSent = writeSerial(to[d]->fd, buf, n);
			else 
				nSent = sendto(to[d]->fd, buf, n, 0,
				    (struct sockaddr *)&(to[d]->dst), to[d]->len);

			if (nSent != n) {
				fprintf(stderr,
				    "%s:%d: sendto: %lu != %lu: %s\n",
				    __FILE__, __LINE__, nSent, n,
				    strerror(errno));
				close(to[d]->fd);
				free(to[d]);
				to[d] = NULL;
			}
		} /* for */
	} /* while */

	for (d = 0; d < nDst; d++) {
		close(to[d]->fd);
		free(to[d]);
	}
	close(fd);
	exit(0);
} /* main */
