/*
 * Copyright (c) 2003-2011 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/**
 ** proxy for DGPS connections
 **
 ** listens on one port for sent corrections, and on another one for
 ** proxying requests
 ** re-emit received corrections to hosts that requested them.
 **/

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <err.h>
#include <errno.h>
#include <netdb.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "network.h"

#define MAX_CLIENTS 10

struct sockaddr_in *clients[MAX_CLIENTS];
int verbose = 0;

static void
usage(const char *progname)
{
	fprintf(stderr, "proxy for GPS corrections\n\n");
	fprintf(stderr, "usage: %s [-v]\n", progname);
	fprintf(stderr, "\t-v\tset verbose mode\n");
	exit(1);
}


static void
register_client(int fd)
{
	int i;
	ssize_t n;
	socklen_t len = sizeof(struct sockaddr_in);
	char buf[80];
	struct sockaddr_in *s;
	struct sockaddr_in from;

	/* Read data from client */
	n = recvfrom(fd, buf, sizeof(buf), 0,
	    (struct sockaddr *)&from, &len);
	if (n < 0) {
		warn("%s:%d: recvfrom", __func__, __LINE__);
		return;
	}
	if (verbose > 1) {
		printf("%s: got %ld bytes from %s\n", __func__, (long)n,
		    inet_ntoa(from.sin_addr));
	}
	/* filter bogus clients */
	if (strcmp(buf, DGPS_PROXY_HELLO) != 0) {
		if (verbose > 1)
			printf("%s: ignoring invalid hello\n", __func__);
		return;
	}
	/* filter duplicate clients */
	for (i = 0; i < MAX_CLIENTS; i++) {
		if (clients[i] != NULL 
		    && memcmp(&clients[i]->sin_addr, 
			&from.sin_addr, sizeof(struct in_addr)) == 0) {
			/* already registered */
			if (verbose > 1)
				printf("%s: updating registration for %s\n",
				    __func__, inet_ntoa(from.sin_addr));
			/* Update port number on client side */
			clients[i]->sin_port = from.sin_port;
			return;
		}
	}
	/* look for a free slot */
	for (i = 0; i < MAX_CLIENTS; i++) 
		if (clients[i] == NULL) 
			break;
	if (i == MAX_CLIENTS) {
		/* client table full */
		warnx("%s: client table full\n", __func__);
		return;
	}
	/* register new client */
	s = malloc(len);
	memcpy(s, &from, len);
	clients[i] = s;
	if (verbose) 
		printf("\tregistered client %s\n", inet_ntoa(from.sin_addr));
	return;
}

static int
forward_to_client(char *buf, ssize_t n, int outFd, struct sockaddr_in *to)
{
	ssize_t nSent;
	socklen_t len = sizeof(struct sockaddr_in);

	if (verbose) 
		printf("\tforwarding to %s\n", inet_ntoa(to->sin_addr));
	nSent = sendto(outFd, buf, n, 0,
	    (struct sockaddr *)to, len); /* XXX */
	if (nSent != n) {
		warn("%s: sendto", __func__);
	}
	return nSent;
}

static void
handle_data(int dataFd, int proxyFd)
{
	char buf[1500];
	struct sockaddr_in from;
	socklen_t len = sizeof(struct sockaddr_in);
	ssize_t n;
	int i;

	n = recvfrom(dataFd, buf, sizeof(buf), 0,
	    (struct sockaddr *)&from, &len);
	if (n <= 0)
		warn("recvfrom(dataFd)");
	if (verbose > 1)
		printf("handle_data: got %lu bytes from %s\n", 
		    n, inet_ntoa(from.sin_addr));
	/* send corrections datagram */
	for (i = 0; i < MAX_CLIENTS; i++) {
		if (clients[i] != NULL &&
		    forward_to_client(buf, n, proxyFd, 
			clients[i]) != n) {
			fprintf(stderr, "%s disconnected\n",
			    inet_ntoa(clients[i]->sin_addr));
			free(clients[i]);
			clients[i] = NULL;
		}
	} /* for */
}

int
main(int argc, char *argv[])
{
	struct pollfd fds[MAX_CLIENTS + 2];
	char *prog = argv[0];
	int ch, dataFd, proxyFd, nfd;

	while ((ch = getopt(argc, argv, "v")) != -1) {
		switch (ch) {
		case 'v':
			verbose++;
			break;
		case '?':
		default:
			usage(prog);
		} /* switch */
	}
	argc -= optind;
	argv += optind;

	dataFd = setupRcvSocket(GPS_PORT);
	if (dataFd < 0)
		err(2, "setup data listener");
	proxyFd = setupRcvSocket(GPS_PROXY_PORT);
	if (proxyFd < 0)
		err(2, "setup proxy listener");

	while (1) {
		fds[0].fd = dataFd;
		fds[0].events = POLLIN;
		fds[1].fd = proxyFd;
		fds[1].events = POLLIN;
		nfd = poll(fds, 2, -1);
		/* Check for errors */
		if (nfd == -1)
			errx(1, "poll error");
		/* client connection ? */
		if (fds[1].revents & POLLIN) {
			register_client(proxyFd);
		}
		/* data received ? */
		if (fds[0].revents & POLLIN)
			handle_data(dataFd, proxyFd);

	} /* while */
}
